import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET",
    format=FORMAT,
    datefmt="[%X]",
    handlers=[
        RichHandler(
            markup=True,
            keywords=["README.md", "info.json", "primer.bed", "reference.fasta"],
        ),
    ],
)

log = logging.getLogger("rich")



import unittest
from copy import deepcopy

from primal_page.schemas import (
    BedfileVersion,
    Collection,
    Info,
    PrimerClass,
    SchemeStatus,
)

base_info = Info(
    ampliconsize=400,
    schemeversion="v1.0.0",
    schemename="test",
    primer_bed_md5="hello",
    reference_fasta_md5="world",
    status=SchemeStatus.DRAFT,
    citations=set("nobel-prize"),
    authors=["artic", "developer"],
    algorithmversion="test",
    species=set("sars-cov-2"),
    articbedversion=BedfileVersion.V3,
    collections=set([Collection.WHOLE_GENOME]),
    primerclass=PrimerClass.PRIMERSCHEMES,
)


class TestAddAuthor(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_author_add_append(self):
        """Test adding an author to the Info object"""

        local_info = deepcopy(self.info)
        new_author = "test-author"

        # Check the author is not present
        self.assertNotIn(new_author, local_info.authors)

        # Append the author
        local_info.author_add(new_author, None)

        # Check the author is present at the end
        self.assertEqual(local_info.authors[-1], new_author)

    def test_author_add_insert(self):
        """Test adding an author to the Info object"""

        local_info = deepcopy(self.info)
        new_author = "test-author"
        new_index = 1

        # Check the author is not present
        self.assertNotIn(new_author, local_info.authors)

        # Insert the author
        local_info.author_add(new_author, new_index)

        # Check the author is present at the index
        self.assertEqual(local_info.authors[new_index], new_author)

    def test_author_add_invalid_index(self):
        """Test adding an author to the Info object with an invalid index"""

        local_info = deepcopy(self.info)
        new_author = "test-author"
        new_index = len(local_info.authors) + 1

        # Check the author is not present
        self.assertNotIn(new_author, local_info.authors)

        # Insert the author
        local_info.author_add(new_author, new_index)

        # Check the author is present at the end
        self.assertEqual(local_info.authors[-1], new_author)


class TestRemoveAuthor(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_author_remove_valid(self):
        """Test removing an author from the Info object"""

        local_info = deepcopy(self.info)
        author_to_remove = local_info.authors[0]

        # Check the author is present
        self.assertIn(author_to_remove, local_info.authors)

        # Remove the author
        local_info.author_remove(author_to_remove)

        # Check the author is not present
        self.assertNotIn(author_to_remove, local_info.authors)

    def test_author_remove_invalid(self):
        """Test removing an author from the Info object"""

        local_info = deepcopy(self.info)
        author_to_remove = "invalid-author"

        # Check the author is not present
        self.assertNotIn(author_to_remove, local_info.authors)

        # Remove the author
        with self.assertRaises(ValueError):
            local_info.author_remove(author_to_remove)


class TestReorderAuthors(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_author_reorders_valid(self):
        """Test reordering authors in the Info object"""

        local_info = deepcopy(self.info)
        new_order = [1, 0]

        # Check the authors are in the correct order
        self.assertEqual(local_info.authors, ["artic", "developer"])

        # Reorder the authors
        local_info.author_reorders(new_order)

        # Check the authors are in the correct order
        self.assertEqual(local_info.authors, ["developer", "artic"])

    def test_author_reorders_invalid_index(self):
        """Test reordering authors in the Info object with an invalid index"""

        local_info = deepcopy(self.info)
        new_order = [1, 2]

        # Check the authors are in the correct order
        self.assertEqual(local_info.authors, ["artic", "developer"])

        # Reorder the authors
        with self.assertRaises(IndexError):
            local_info.author_reorders(new_order)

    def test_author_reorders_duplicate_index(self):
        """Test reordering authors in the Info object with a duplicate index"""

        local_info = deepcopy(self.info)
        new_order = [1, 1]

        # Check the authors are in the correct order
        self.assertEqual(local_info.authors, ["artic", "developer"])

        # Reorder the authors
        with self.assertRaises(ValueError):
            local_info.author_reorders(new_order)

    def test_no_authors_lost(self):
        """Test reordering authors in the Info object with a duplicate index"""

        local_info = deepcopy(self.info)
        new_order = [1]

        # Check the authors are in the correct order
        self.assertEqual(local_info.authors, ["artic", "developer"])

        # Reorder the authors
        local_info.author_reorders(new_order)

        # Check the authors are in the correct order
        self.assertEqual(local_info.authors, ["developer", "artic"])


class TestAddCitation(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_citation_add(self):
        """Test adding a citation to the Info object"""

        local_info = deepcopy(self.info)
        new_citation = "test-citation"

        # Check the citation is not present
        self.assertNotIn(new_citation, local_info.citations)

        # Append the citation
        local_info.citation_add(new_citation)

        # Check the citation is present at the end
        self.assertIn(new_citation, local_info.citations)


class TestRemoveCitation(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_citation_remove_valid(self):
        """Test removing a citation from the Info object"""

        local_info = deepcopy(self.info)
        citation_to_remove = next(iter(local_info.citations))

        # Check the citation is present
        self.assertIn(citation_to_remove, local_info.citations)

        # Remove the citation
        local_info.citation_remove(citation_to_remove)

        # Check the citation is not present
        self.assertNotIn(citation_to_remove, local_info.citations)

    def test_citation_remove_invalid(self):
        """Test removing a citation from the Info object"""

        local_info = deepcopy(self.info)
        citation_to_remove = "invalid-citation"

        # Check the citation is not present
        self.assertNotIn(citation_to_remove, local_info.citations)

        # Remove the citation
        with self.assertRaises(KeyError):
            local_info.citation_remove(citation_to_remove)


class TestAddCollection(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_collection_add(self):
        """Test adding a collection to the Info object"""

        local_info = deepcopy(self.info)
        new_collection = Collection.ARTIC

        # Check the collection is not present
        self.assertNotIn(new_collection, local_info.collections)

        # Append the collection
        local_info.collection_add(new_collection)

        # Check the collection is present at the end
        self.assertIn(new_collection, local_info.collections)


class TestRemoveCollection(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_collection_remove_valid(self):
        """Test removing a collection from the Info object"""

        local_info = deepcopy(self.info)
        collection_to_remove = next(iter(local_info.collections))

        # Check the collection is present
        self.assertIn(collection_to_remove, local_info.collections)

        # Remove the collection
        local_info.collection_remove(collection_to_remove)

        # Check the collection is not present
        self.assertNotIn(collection_to_remove, local_info.collections)

    def test_collection_remove_invalid(self):
        """Test removing a collection from the Info object"""

        local_info = deepcopy(self.info)
        collection_to_remove = Collection.MULTI_TARGET

        # Check the collection is not present
        self.assertNotIn(collection_to_remove, local_info.collections)

        # Remove the collection
        with self.assertRaises(KeyError):
            local_info.collection_remove(collection_to_remove)


class TestChangeDescription(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_change_description(self):
        """Test changing the description of the Info object"""

        local_info = deepcopy(self.info)
        new_description = "test-description"

        # Check the description is not present
        self.assertNotEqual(new_description, local_info.description)

        # Change the description
        local_info.description = new_description

        # Check the description is present
        self.assertEqual(new_description, local_info.description)


class TestChangeStatus(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_change_status(self):
        """Test changing the status of the Info object"""

        local_info = deepcopy(self.info)
        new_status = SchemeStatus.VALIDATED

        # Check the status is not present
        self.assertNotEqual(new_status, local_info.status)

        # Change the status
        local_info.status = new_status

        # Check the status is present
        self.assertEqual(new_status, local_info.status)


class TestChangeLicense(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_change_license(self):
        """Test changing the license of the Info object"""

        local_info = deepcopy(self.info)
        new_license = "test-license"

        # Check the license is not present
        self.assertNotEqual(new_license, local_info.license)

        # Change the license
        local_info.license = new_license

        # Check the license is present
        self.assertEqual(new_license, local_info.license)


class TestChangeDerivedFrom(unittest.TestCase):
    def setUp(self) -> None:
        self.info = base_info

    def test_change_derivedfrom(self):
        """Test changing the derivedfrom of the Info object"""

        local_info = deepcopy(self.info)
        new_derivedfrom = "test-derivedfrom"

        # Check the derivedfrom is not present
        self.assertNotEqual(new_derivedfrom, local_info.derivedfrom)

        # Change the derivedfrom
        local_info.derivedfrom = new_derivedfrom

        # Check the derivedfrom is present
        self.assertEqual(new_derivedfrom, local_info.derivedfrom)


if __name__ == "__main__":
    unittest.main()



import pathlib

import typer
from Bio import SeqIO
from click import UsageError
from primalbedtools.bedfiles import BedLineParser, PrimerNameVersion, group_primer_pairs
from typing_extensions import Annotated

from primal_page.logging import log
from primal_page.modify import hash_file
from primal_page.schemas import Info, validate_ref_select_file

app = typer.Typer(no_args_is_help=True)


def validate_name(infopath: pathlib.Path):
    """
    Validate the schemename, ampliconsize, and schemeversion in the path, ReadME.md, and info.json
    :raises ValueError: If a mismatch is found
    :raises FileNotFoundError: If the ReadME.md does not exist
    """

    # Read in the info.json
    info = Info.model_validate_json(infopath.read_text())
    info_scheme_path = (
        info.schemename + "/" + str(info.ampliconsize) + "/" + info.schemeversion
    )

    # Check the ReadME.md
    readme = infopath.parent / "README.md"
    if not readme.exists():
        raise FileNotFoundError(f"{readme} does not exist")

    # Check the info version matches path version
    version_path = infopath.parent.name
    if info.schemeversion != version_path:
        raise ValueError(
            f"Version mismatch for {info_scheme_path}: info ({info.schemeversion}) != path ({version_path})"
        )

    # Check the amplicon size matches the schemepath
    ampliconsize_path = infopath.parent.parent.name
    if info.ampliconsize != int(ampliconsize_path):
        raise ValueError(
            f"Ampliconsize mismatch for {info_scheme_path}: info ({info.ampliconsize}) != path ({ampliconsize_path})"
        )

    # Check the schemepath matches the path
    schemeid_path = infopath.parent.parent.parent.name
    if info.schemename != schemeid_path:
        raise ValueError(
            f"Schemename mismatch for {info_scheme_path}: info ({info.schemename}) != path ({schemeid_path})"
        )

    # Check the readme has been updated
    readme = readme.read_text()
    if readme.find(info.schemename) == -1:
        raise ValueError(
            f"Scheme name ({info.schemename}) not found in {readme}: {info.schemename}"
        )
    if readme.find(str(info.ampliconsize)) == -1:
        raise ValueError(
            f"Amplicon size ({info.ampliconsize}) not found in {readme}: {info.ampliconsize}"
        )
    if readme.find(info.schemeversion) == -1:
        raise ValueError(
            f"Scheme version ({info.schemeversion}) not found in {readme}: {info.schemeversion}"
        )


def validate_bedfile(bedpath: pathlib.Path, strict: bool = True):
    """
    Uses primalbedtools to validate the bedfiles.
    :raises ValueError: If the bedfile contains old primer names
    """
    _header, bedlines = BedLineParser.from_file(bedpath)

    for bedline in bedlines:
        if bedline.primername_version != PrimerNameVersion.V2:
            raise ValueError(
                f"Bedfile {bedpath} contains old primer names ({bedline.primername})"
            )

    # Carry out strict checks
    if not strict:
        return

    # Look for both left and right primers
    primer_pairs = group_primer_pairs(bedlines)
    for fkmers, rkmers in primer_pairs:
        if not fkmers:
            raise ValueError(
                f"Missing forward primer for {rkmers[0].amplicon_prefix}_{rkmers[0].amplicon_number} in {bedpath}"
            )
        if not rkmers:
            raise ValueError(
                f"Missing reverse primer for {fkmers[0].amplicon_prefix}_{fkmers[0].amplicon_number} in {bedpath}"
            )


def validate_hashes(infopath: pathlib.Path):
    """
    Reads in the info.json and checks the hashes for ./primer.bed and ./reference.fasta
    """
    # Read in the info.json
    info = Info.model_validate_json(infopath.read_text())
    info_scheme_path = info.get_schemepath()

    # Check the hashes
    if info.primer_bed_md5 != hash_file(infopath.parent / "primer.bed"):
        raise ValueError(
            f"MD5 mismatch for {info_scheme_path}:primer.bed: info ({info.primer_bed_md5}) != file ({hash_file(infopath.parent / 'primer.bed')})"
        )

    if info.reference_fasta_md5 != hash_file(infopath.parent / "reference.fasta"):
        raise ValueError(
            f"MD5 mismatch for {info_scheme_path}:reference.fasta: info ({info.reference_fasta_md5}) != file ({hash_file(infopath.parent / 'reference.fasta')})"
        )


def validate_ref_and_bed(infopath: pathlib.Path):
    """
    Checks the chrom in bedfile and reference.fasta file.
    """
    # Read in the info.json
    info = Info.model_validate_json(infopath.read_text())
    info_scheme_path = info.get_schemepath()

    # Read in the bedfile
    _header, bedlines = BedLineParser.from_file(infopath.parent / "primer.bed")
    chrom_names = {bedline.chrom for bedline in bedlines}

    # Read in the reference.fasta
    ref_index = SeqIO.index(infopath.parent / "reference.fasta", "fasta")

    # Look for chroms in the bedfile that are not in the reference.fasta
    delta = chrom_names - set(ref_index.keys())
    if chrom_names - set(ref_index.keys()):
        raise ValueError(
            f"{info_scheme_path}: chroms in primer.bed that are not in reference.fasta: {delta}"
        )
    # Look for chroms in the reference.fasta that are not in the bedfile
    delta = set(ref_index.keys()) - chrom_names
    if delta:
        raise ValueError(
            f"{info_scheme_path}: chroms in reference.fasta that are not in primer.bed: {delta}"
        )


def validate_ref_selection(infopath: pathlib.Path):
    # Read in the info.json
    info = Info.model_validate_json(infopath.read_text())
    info_scheme_path = info.get_schemepath()

    # Check the ref-selection
    if info.refselect is None:
        return

    for chrom, data in info.refselect.items():
        file_path = infopath.parent / data["filename"]
        # Check the file exists
        if not file_path.is_file():
            raise FileNotFoundError(f"{file_path} is not a file")

        # Check the md5
        file_md5 = hash_file(file_path)
        if data["md5"] != file_md5:
            raise ValueError(
                f"MD5 mismatch for {info_scheme_path}:{chrom}: info ({data['md5']}) != file ({file_md5})"
            )
        # Revalidate the ref-select file
        validate_ref_select_file(
            info=info, chrom=chrom, ref_select=file_path, infopath=infopath
        )


def validate_ref_select_final(primerschemes_dir: pathlib.Path):
    """
    Checks that the clades in the ref-select files exist
    """

    for info_path in primerschemes_dir.rglob("*info.json"):
        info = Info.model_validate_json(info_path.read_text())
        info_scheme_path = info.get_schemepath()

        if info.refselect is None:
            continue
        else:
            log.info(f"[blue]Checking[/blue]:\t{info_scheme_path}")

        for _chrom, data in info.refselect.items():
            file_path = info_path.parent / data["filename"]
            # Check the file exists
            if not file_path.is_file():
                raise FileNotFoundError(f"{file_path} is not a file")

            # Read in the fasta headers
            with open(file_path) as fasta_file:
                headers = [line.split() for line in fasta_file if line.startswith(">")]

            alt_refs = {header[0].strip(): header[1].strip() for header in headers}

            # Check the alt-refs schemes exist
            version_dir = info_path.parent.parent
            for _newchrom, clade in alt_refs.items():
                clade_path = version_dir / f"{info.schemeversion}-{clade}"

                if not clade_path.exists():
                    raise FileNotFoundError(f"Could not find scheme {clade_path}")

                log.debug(f"Found:\t{'/'.join(clade_path.parts[-3:])}")


@app.command(no_args_is_help=True)
def scheme(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json",
            readable=True,
            exists=True,
        ),
    ],
):
    """
    Validate a single scheme
    """
    # Get the schemename
    info = Info.model_validate_json(schemeinfo.read_text())
    info_scheme_path = info.get_schemepath()

    errors = []

    try:
        validate_name(schemeinfo)
    except Exception as e:
        errors.append(str(e))
    try:
        bedfile = schemeinfo.parent / "primer.bed"
        validate_bedfile(bedfile)
    except Exception as e:
        errors.append(str(e))
    try:
        validate_hashes(schemeinfo)
    except Exception as e:
        errors.append(str(e))
    try:
        validate_ref_selection(schemeinfo)
    except Exception as e:
        errors.append(str(e))
    try:
        validate_ref_and_bed(schemeinfo)
    except Exception as e:
        errors.append(str(e))

    if errors:
        log.error(f"[red]Errored[/red]:\t[blue]{info_scheme_path}[/blue]")
        raise UsageError(message="\n".join(errors))
    else:
        log.info(f"[green]Success[/green]:\t[blue]{info_scheme_path}[/blue]")


@app.command(no_args_is_help=True)
def all_schemes(
    directory: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to the directory",
            readable=True,
            exists=True,
            dir_okay=True,
        ),
    ],
):
    """
    Validate all schemes in a directory. Calls the scheme command for each scheme and checks for final ref-select files.
    """
    errors = []
    for schemeinfo in directory.rglob("*/info.json"):
        try:
            scheme(schemeinfo)
        except Exception as e:
            errors.append(str(e))

    # Check for final ref-select files
    try:
        validate_ref_select_final(directory)
    except Exception as e:
        errors.append(str(e))

    if errors:
        raise UsageError(message="\n".join(errors))



import pathlib
import unittest

from primal_page.validate import validate_bedfile, validate_hashes, validate_name

# Everything is correct in this
valid_info_path = pathlib.Path(
    "tests/test_input/validate/valid-scheme/400/v5.4.2/info.json"
)
# All hashes are incorrect
invalid_info_path = pathlib.Path(
    "tests/test_input/validate/invalid-scheme/400/v5.4.2/info.json"
)


class TestValidateHashes(unittest.TestCase):
    def test_validate_hashes(self):
        """
        Checks the hashes in the valid info.json
        """
        validate_hashes(valid_info_path)

    def test_validate_hashes_invalid(self):
        """
        Checks the hashes in the invalid info.json
        """
        with self.assertRaises(ValueError):
            validate_hashes(invalid_info_path)


class TestValidateName(unittest.TestCase):
    def test_validate_name(self):
        """
        Checks the name in the valid info.json
        """
        validate_name(valid_info_path)

    def test_validate_name_invalid(self):
        """
        Checks the name in the invalid info.json
        """
        with self.assertRaises((ValueError, FileNotFoundError)):
            validate_hashes(invalid_info_path)


class TestValidateBedFile(unittest.TestCase):
    def test_validate_bedfile(self):
        """
        Checks the bedfile in the valid info.json
        """
        validate_bedfile(valid_info_path.parent / "primer.bed")

    def test_validate_bedfile_invalid(self):
        """
        Checks the bedfile in the invalid info.json
        """
        with self.assertRaises((ValueError, IndexError)):
            validate_bedfile(invalid_info_path.parent / "primer.bed")



import unittest

from primal_page.schemas import Links


class TestLinks(unittest.TestCase):
    def test_link_add(self):
        links = Links()

        # Add a link to all correct linkfield
        linktoadd = "https://example.com"
        for linkfield in links.model_fields.keys():
            links.append_link(linkfield, linktoadd)
            self.assertEqual(getattr(links, linkfield), [linktoadd])

        # Add a link to an incorrect linkfield
        with self.assertRaises(AttributeError):
            links.append_link("invalid", linktoadd)

    def test_link_remove(self):
        linktoremove = "https://example.com"
        links = Links(protocols=[linktoremove])

        # Remove the link from protocols
        links.link_remove("protocols", linktoremove)
        self.assertEqual(links.protocols, [])

        # Remove a link from an incorrect linkfield
        with self.assertRaises(AttributeError):
            links.link_remove("invalid", linktoremove)

        # Remove a link that is not in the linkfield
        with self.assertRaises(ValueError):
            links.link_remove("protocols", linktoremove)


if __name__ == "__main__":
    unittest.main()



import hashlib
import pathlib
import unittest

import requests

from primal_page.download import fetch_index, validate_hashes


class TestValidateHashes(unittest.TestCase):
    def setUp(self) -> None:
        self.outdir = pathlib.Path("tests/test_output")
        self.outdir.mkdir(exist_ok=True)

    def test_invalid_hash(self):
        """
        Ensures that if the hashes do not match, the file is not written.
        """
        outfile = self.outdir / "no_file_should_write.txt"
        outfile.unlink(missing_ok=True)

        with self.assertRaises(ValueError):
            validate_hashes(
                "test",
                "invalid",
                outfile,
            )

        # Check the file was not written
        ## THIS IS THE MOST IMPORTANT PART OF THE TEST.
        self.assertFalse(outfile.exists())

    def test_valid_hash(self):
        """
        If the hashes match, the file should be written.
        """
        outfile = self.outdir / "should_write.txt"
        outfile.unlink(missing_ok=True)

        text = "test"

        # Create the hash
        expected_hash = hashlib.md5(text.encode()).hexdigest()

        # Run the function
        validate_hashes(text, expected_hash, outfile)

        # Check the file was written
        self.assertTrue(outfile.exists())

        # Check the file contents
        self.assertEqual(outfile.read_text(), text)

        # Remove the file
        outfile.unlink()


class TestFetchIndex(unittest.TestCase):
    def setUp(self) -> None:
        self.index_url = (
            "https://raw.githubusercontent.com/quick-lab/primerschemes/main/index.json"
        )

    def test_fetch_index(self):
        """
        Ensures that the index is fetched correctly.
        """
        from primal_page.download import fetch_index

        index = fetch_index(self.index_url)

        self.assertIsInstance(index, dict)
        self.assertIn("primerschemes", index)

    def test_invalid_url(self):
        """
        Ensures that an invalid URL raises a ValueError.
        """

        with self.assertRaises(requests.exceptions.HTTPError):
            fetch_index(
                "https://raw.githubusercontent.com/quick-lab/primerschemes/main/invalid.json"
            )


if __name__ == "__main__":
    unittest.main()



import json
import pathlib
import shutil
from importlib.metadata import version
from typing import Annotated, Optional

import typer
from Bio import Seq, SeqIO, SeqRecord
from primalbedtools.bedfiles import BedFileModifier, BedLineParser, PrimerNameVersion

from primal_page.aliases import app as aliases_app
from primal_page.bedfiles import BedfileVersion
from primal_page.build_index import create_index
from primal_page.dev import app as dev_app
from primal_page.download import app as download_app
from primal_page.errors import InvalidReference, SchemeExists
from primal_page.logging import log
from primal_page.modify import app as modify_app
from primal_page.modify import generate_files, hash_file
from primal_page.schemas import (
    Collection,
    Info,
    IUPACAmbiguousDNA,
    Links,
    PrimerClass,
    SchemeStatus,
)
from primal_page.validate import app as validate_app
from primal_page.validate import validate_bedfile, validate_ref_select_file

# Create the typer app
app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)
app.add_typer(
    modify_app,
    name="modify",
    help="Modify an existing scheme's metadata (info.json)",
)
app.add_typer(
    download_app,
    name="download",
    help="Download schemes from the index.json",
)
app.add_typer(dev_app, name="dev", help="Development commands", hidden=True)
app.add_typer(aliases_app, name="aliases", help="Manage aliases")
app.add_typer(validate_app, name="validate", help="Validate a scheme")


def typer_callback_version(value: bool):
    if value:
        typer.echo(f"primal-page version: {version('primal-page')}")
        raise typer.Exit()


@app.callback()
def primal_page(
    value: Annotated[bool, typer.Option] = typer.Option(
        False, "--version", callback=typer_callback_version
    ),
):
    pass


def validate_ref_records(ref_records: list[SeqRecord.SeqRecord]):
    """
    Validate the reference.fasta file
    :param ref_file: The path to the reference.fasta file
    :raises InvalidReference: If the reference.fasta file is invalid
    """

    for record in ref_records:
        assert isinstance(record.seq, Seq.Seq)
        # Check DNA sequence
        seq_bases = set(record.seq)
        if not seq_bases.issubset(IUPACAmbiguousDNA):
            raise InvalidReference(
                f"Invalid DNA bases ({', '.join(seq_bases.difference(IUPACAmbiguousDNA))}) found in {record.id}"
            )
        if len(record.seq) == 0:
            raise InvalidReference(f"Empty sequence found in {record.id}")


@app.command(no_args_is_help=True)
def create(
    schemename: Annotated[
        str,
        typer.Option(help="The name of the scheme"),
    ],
    ampliconsize: Annotated[
        int,
        typer.Option(help="Amplicon size", min=100),
    ],
    schemeversion: Annotated[
        str,
        typer.Option(
            help="Scheme version, default is parsed from config.json",
        ),
    ],
    species: Annotated[
        list[int],
        typer.Option(
            help="The species this scheme targets. Please use NCBI taxonomy ids"
        ),
    ],
    authors: Annotated[
        list[str],
        typer.Option(
            help="Any authors. To provide multiple, use --authors '1' --authors '2'"
        ),
    ],
    primerbed: Annotated[
        pathlib.Path,
        typer.Option(
            help="The path to the primer.bed file",
            readable=True,
        ),
    ],
    reference: Annotated[
        pathlib.Path,
        typer.Option(
            help="The path to the reference.fasta file",
            readable=True,
        ),
    ],
    schemestatus: Annotated[
        SchemeStatus, typer.Option(help="Scheme status")
    ] = SchemeStatus.DRAFT.value,  # type: ignore
    citations: Annotated[
        list[str], typer.Option(help="Any associated citations. Please use DOI")
    ] = [],
    output: Annotated[
        pathlib.Path,
        typer.Option(help="Where to output the scheme", writable=True),
    ] = pathlib.Path("primerschemes"),
    configpath: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Where the config.json file is located", readable=True, exists=True
        ),
    ] = None,
    algorithmversion: Annotated[
        Optional[str], typer.Option(help="The version of primalscheme or other")
    ] = None,
    description: Annotated[
        Optional[str], typer.Option(help="A description of the scheme")
    ] = None,
    derivedfrom: Annotated[
        Optional[str],
        typer.Option(help="Which scheme has this scheme been derived from"),
    ] = None,
    primerclass: Annotated[
        PrimerClass, typer.Option(help="The primer class")
    ] = PrimerClass.PRIMERSCHEMES.value,  # type: ignore
    collection: Annotated[
        Optional[list[Collection]],
        typer.Option(
            help="The collection tags. To provide multiple, use --collection '1' --collection '2'"
        ),
    ] = None,
    link_protocol: Annotated[
        list[str], typer.Option(help="Optional link to protocol")
    ] = [],
    link_validation: Annotated[
        list[str], typer.Option(help="Optional link to validation data")
    ] = [],
    links_homepage: Annotated[
        list[str], typer.Option(help="Optional link to homepage")
    ] = [],
    link_vendor: Annotated[
        list[str], typer.Option(help="Optional link to vendors")
    ] = [],
    link_misc: Annotated[
        list[str], typer.Option(help="Optional miscellaneous link")
    ] = [],
    fix: Annotated[bool, typer.Option(help="Attempt to fix the scheme")] = False,
    contact_info: Annotated[
        Optional[str], typer.Option(help="Contact information")
    ] = None,
    additional_files: Annotated[
        list[pathlib.Path],
        typer.Option(help="Additional files to include in the ./work directory"),
    ] = [],
    ref_select: Annotated[
        Optional[tuple[str, pathlib.Path]],
        typer.Option(help="Reference selection file. In the form 'chromosome file'"),
    ] = None,
):
    """Create a new scheme in the required format"""

    # Parse the primer.bed file
    try:
        headers, bedlines = BedLineParser().from_file(primerbed)
    except ValueError as e:
        raise typer.BadParameter(f"Error parsing {primerbed}: {e}") from e
    bedlines = BedFileModifier.sort_bedlines(bedlines)

    # get all primername versions
    primer_name_versions_to_name: dict[PrimerNameVersion, str] = {
        line.primername_version: line.primername for line in bedlines
    }

    if (
        len(primer_name_versions_to_name) > 1
        or PrimerNameVersion.V2 not in primer_name_versions_to_name
    ) and not fix:
        raise typer.BadParameter(
            f"Primernames ({[v for k, v in primer_name_versions_to_name.items() if k != PrimerNameVersion.V2]})"
            " are not in the correct format. Please update or try with --fix to attempt to parse."
        )

    if fix:
        # Attempt to fix the primer.bed file
        bedlines = BedFileModifier.update_primernames(bedlines)

    # Find the reference.fasta file
    reference_records = list(SeqIO.parse(reference, "fasta"))

    if len(reference_records) == 0:
        raise InvalidReference(f"Empty reference file {reference}")

    for record in reference_records:
        record.seq = record.seq.upper()

    # Validate the reference.fasta file
    validate_ref_records(reference_records)

    # Search for config.json
    config_json: None | dict = None
    if configpath is not None:
        config_json = json.load(configpath.open())
        assert isinstance(config_json, dict)
        # Remove some paths from the config
        config_json.pop("output_dir", None)
        # These hashes are no longer used, so remove to prevent confusion
        md5s_keys_to_remove = {k for k in config_json.keys() if k.endswith("md5")}
        for k in md5s_keys_to_remove:
            config_json.pop(k)

    # Multiple pngs/htmls/msas are allowed
    # The single check is mainly to prevent multiple schemes via providing the wrong directory
    # At this point we know we have a single scheme, due to having a single primer.bed, reference.fasta, and config.json

    pngs = [path for path in additional_files if path.name.endswith(".png")]

    # Create the collections set
    collections = {x for x in collection} if collection is not None else set()

    # Create the links set
    links = Links(
        protocols=link_protocol,
        validation=link_validation,
        homepage=links_homepage,
        vendors=link_vendor,
        misc=link_misc,
    )
    # Create the info.json
    # Generate the md5s
    info = Info(
        ampliconsize=ampliconsize,
        schemeversion=schemeversion,
        schemename=schemename,
        primer_bed_md5="NONE",  # Will be updated later
        reference_fasta_md5="NONE",  # Will be updated later
        status=schemestatus,
        citations=set(citations),
        authors=authors,
        algorithmversion=algorithmversion,  # type: ignore
        species=set(species),
        description=description,
        derivedfrom=derivedfrom,
        primerclass=primerclass,
        articbedversion=BedfileVersion.V3,
        collections=collections,
        links=links,
        contactinfo=contact_info,
    )

    # Parse ref_select
    if ref_select is not None:
        validate_ref_select_file(info, ref_select[0], ref_select[1], primerbed)

    #####################################
    # Final validation and create files #
    #####################################
    # Everything needs to be validated before creating files

    # Check if the repo already exists
    repo_dir = output / schemename / str(ampliconsize) / schemeversion
    if repo_dir.exists():
        raise SchemeExists(f"{repo_dir} already exists")
    repo_dir.mkdir(parents=True)

    # If this fails it will deleted the half completed scheme
    # Need to check the repo doesn't already exist
    try:
        # Write the primer.bed file
        BedLineParser().to_file(repo_dir / "primer.bed", headers, bedlines)

        # validate the bedfile
        # let the exception bubble up
        validate_bedfile(repo_dir / "primer.bed")

        # Write the reference.fasta file records
        with open(repo_dir / "reference.fasta", "w") as ref_file:
            SeqIO.write(reference_records, ref_file, "fasta")

        # Add the reference selection files
        if ref_select is not None:
            info.add_refselect(ref_select[1], ref_select[0], repo_dir / "info.json")

        # Update the hashes in the info.json
        info.primer_bed_md5 = hash_file(repo_dir / "primer.bed")
        info.reference_fasta_md5 = hash_file(repo_dir / "reference.fasta")

        working_dir = repo_dir / "work"
        working_dir.mkdir()

        if config_json is not None:
            # Write out the config.json
            with open(working_dir / "config.json", "w") as configfile:
                json.dump(config_json, configfile, indent=4, sort_keys=True)

        # Copy over misc files
        for file in additional_files:
            shutil.copy(file, working_dir / file.name)

        # Write out the info.json and readme
        generate_files(info, repo_dir, pngs=pngs)

    except Exception as e:
        # Cleanup
        shutil.rmtree(repo_dir)
        raise typer.BadParameter(f"{e}\nCleaning up {repo_dir}") from e


@app.command(no_args_is_help=True)
def build_index(
    gitaccount: Annotated[
        str,
        typer.Option(help="The name of the github account"),
    ] = "quick-lab",
    gitserver: Annotated[
        str,
        typer.Option(help="The name of the github server"),
    ] = "https://github.com/",
    parentdir: Annotated[
        pathlib.Path, typer.Option(help="The parent directory")
    ] = pathlib.Path("."),
    git_commit_sha: Annotated[
        Optional[str], typer.Option(help="The git commit")
    ] = None,
    force: Annotated[
        bool, typer.Option(help="Force the creation of the index.json")
    ] = False,
):
    """Build an index.json file from all schemes in the directory"""

    create_index(gitserver, gitaccount, parentdir, git_commit_sha, force)


@app.command(no_args_is_help=True)
def remove(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
):
    """Remove a scheme's version from the repo, will also remove size and schemename directories if empty"""
    # Check that this is an info.json file (for safety)
    if schemeinfo.name != "info.json":
        raise typer.BadParameter(f"{schemeinfo} is not an info.json file")

    # Remove the schemeversion directory
    shutil.rmtree(schemeinfo.parent)

    # Move up the path and remove the size directory if empty
    size_dir = schemeinfo.parent.parent
    if len(list(size_dir.iterdir())) == 0:
        size_dir.rmdir()

    # Move up the path and remove the schemename directory if empty
    scheme_dir = size_dir.parent
    if len(list(scheme_dir.iterdir())) == 0:
        scheme_dir.rmdir()

    log.info(f"Removed {schemeinfo}")


if __name__ == "__main__":
    app()



import json
import pathlib
import shutil
import unittest

from primal_page.main import create
from primal_page.schemas import SchemeStatus


class TestCreate(unittest.TestCase):
    def setUp(self) -> None:
        # Required IO params
        self.schemepath = pathlib.Path("tests/test_input/test_covid")
        self.output = pathlib.Path("tests/test_output/test_covid")

        # Required params
        self.ampliconsize = 400
        self.schemeversion = "v1.0.0"
        self.species = [10]
        self.schemestatus = SchemeStatus.DRAFT
        self.citations = ["test-citation:124"]
        self.authors = ["artic"]
        self.schemename = "test-data-covid"

        # Parsed / optional params
        self.primerbed = pathlib.Path("tests/test_input/test_covid/primer.bed")
        self.reference = pathlib.Path("tests/test_input/test_covid/reference.fasta")
        self.configpath = pathlib.Path("tests/test_input/test_covid/config.json")
        self.algorithmversion = "primalscheme-test"
        self.description = "test-description"
        self.derivedfrom = "test-derivedfrom"

        # Clean up any previous test runs
        if (self.output / self.schemename).exists():
            shutil.rmtree(self.output / self.schemename)

    def test_create_minimal(self):
        """Test the creation of the scheme, using the config and required params"""

        # Run the create function
        create(
            output=self.output,
            ampliconsize=self.ampliconsize,
            schemeversion=self.schemeversion,
            species=self.species,
            schemestatus=self.schemestatus,
            citations=self.citations,
            authors=self.authors,
            schemename=self.schemename,
            reference=self.reference,
            primerbed=self.primerbed,
            algorithmversion=self.algorithmversion,
        )

        # Check the output files exist
        self.assertTrue(
            (
                self.output
                / self.schemename
                / str(self.ampliconsize)
                / self.schemeversion
            ).exists()
        )

        # Check the info file exists
        self.assertTrue(
            (
                self.output
                / self.schemename
                / str(self.ampliconsize)
                / self.schemeversion
                / "info.json"
            ).exists()
        )

        # Check the details in the info file are correct
        info = json.loads(
            open(
                self.output
                / self.schemename
                / str(self.ampliconsize)
                / self.schemeversion
                / "info.json"
            ).read()
        )
        # Check the main fields are correct
        self.assertEqual(info["schemename"], self.schemename)
        self.assertEqual(info["ampliconsize"], self.ampliconsize)
        self.assertEqual(info["schemeversion"], self.schemeversion)

        # Check the optional fields are correct
        self.assertEqual(info["status"], self.schemestatus.value)
        self.assertEqual(info["authors"], self.authors)
        self.assertEqual(info["citations"], self.citations)
        self.assertEqual(info["species"], self.species)

        # Test the non provided fields are correct
        self.assertEqual(
            info["algorithmversion"], self.algorithmversion
        )  # parsed from config
        self.assertEqual(info["description"], None)
        self.assertEqual(info["derivedfrom"], None)


if __name__ == "__main__":
    unittest.main()



import hashlib
import pathlib
import re
from enum import Enum
from typing import Annotated

from Bio import SeqIO
from primalbedtools.bedfiles import BedLineParser
from pydantic import BaseModel, PositiveInt
from pydantic.functional_validators import AfterValidator

from primal_page.bedfiles import BedfileVersion
from primal_page.errors import InvalidSchemeID, InvalidSchemeName, InvalidSchemeVersion


def hash_file(fname: pathlib.Path) -> str:
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


INFO_SCHEMA = "v2.1.0"

SCHEMENAME_PATTERN = r"^[a-z0-9][a-z0-9-]*[a-z0-9]$"
VERSION_PATTERN = r"^v\d+\.\d+\.\d+(-[a-z0-9]+)?$"
DOI_PATTERN = r"^(https:\/\/|dx.)*(doi.org\/|doi:)10.\d+\/.*$"


class PrimerClass(Enum):
    PRIMERSCHEMES = "primerschemes"


class SchemeStatus(Enum):
    WITHDRAWN = "withdrawn"
    DEPRECATED = "deprecated"
    AUTOGENERATED = "autogenerated"
    DRAFT = "draft"
    TESTING = "tested"
    VALIDATED = "validated"


def validate_scheme_id(schemeid) -> tuple[str, str, str]:
    """
    Parse the schemeid into its components
    :raises InvalidSchemeID: if the schemeid is invalid
    """
    try:
        (
            schemename,
            ampliconsize,
            schemeversion,
        ) = schemeid.split("/")
    except ValueError as e:
        raise InvalidSchemeID(
            f"{schemeid} needs to be in the form (schemename)/(ampliconsize)/(schemeversion)"
        ) from e

    # Validate each part separately
    try:
        schemename = validate_schemename(schemename)
        schemeversion = validate_schemeversion(schemeversion)
        ampliconsize = int(ampliconsize)
    except InvalidSchemeName as e:
        raise InvalidSchemeID(
            f"{schemename} is an invalid schemename in {schemeid}"
        ) from e
    except InvalidSchemeVersion as e:
        raise InvalidSchemeID(
            f"{schemeversion} is an invalid schemeversion in {schemeid}"
        ) from e
    except ValueError as e:
        raise InvalidSchemeID(
            f"{ampliconsize} is an invalid ampliconsize in {schemeid}"
        ) from e

    return schemename, str(ampliconsize), schemeversion


def validate_schemeversion(version: str) -> str:
    """
    Validate the schemeversion. Must be in the form of v(int).(int).(int) with optional suffix -[a-z0-9].
    :raises InvalidSchemeVersion: if the schemeversion is invalid
    """
    if not re.match(VERSION_PATTERN, version):
        raise InvalidSchemeVersion(
            f"Invalid version: ({version}). Must match be in form of v(int).(int).(int) with optional suffix -[a-z0-9]"
        )
    return version


def validate_schemename(schemename: str) -> str:
    """
    Validate the schemename
    :raises InvalidSchemeName: if the schemename is invalid
    """
    if not re.match(SCHEMENAME_PATTERN, schemename):
        raise InvalidSchemeName(
            f"Invalid schemename: ({schemename}). Must only contain a-z, 0-9, and -. Cannot start or end with -"
        )
    return schemename


def not_empty(x: list | set | str) -> list | set | str:
    if len(x) == 0:
        raise ValueError("Cannot be empty")
    return x


IUPACAmbiguousDNA = {
    "A",
    "G",
    "K",
    "Y",
    "B",
    "S",
    "N",
    "H",
    "C",
    "W",
    "D",
    "R",
    "M",
    "T",
    "V",
}


class Collection(Enum):
    # Authors
    ARTIC = "ARTIC"
    MODJADJI = "MODJADJI"
    QUICK_LAB = "QUICK-LAB"
    COMMUNITY = "COMMUNITY"
    # Uses
    WASTE_WATER = "WASTE-WATER"
    CLINICAL_ISOLATES = "CLINICAL-ISOLATES"
    # Coverage types
    WHOLE_GENOME = "WHOLE-GENOME"
    PANEL = "PANEL"
    MULTI_TARGET = "MULTI-TARGET"
    # Other Scheme Generators
    VARVAMP = "VARVAMP"

    def __ord__(self):
        return self.value


class Links(BaseModel):
    protocols: list[str] = []
    validation: list[str] = []
    homepage: list[str] = []
    vendors: list[str] = []
    misc: list[str] = []

    def getattr(self, link_type: str):
        """
        Get the list of links
        Raises:
        -   AttributeError if the link_type is not in the model_fields
        """
        return getattr(self, link_type)

    def append_link(
        self,
        link_type: str,
        link: str,
    ):
        """
        Add a link to the correct list
        Raises:
        -   AttributeError if the link_type is not in the model_fields
        """
        # Append the link to the correct list
        getattr(self, link_type).append(link)

    def link_remove(
        self,
        link_type: str,
        link: str,
    ):
        """
        Remove a the first occurrence of the link from the correct list
        Raises:
        -   ValueError if the link is not in the list
        -   AttributeError if the link_type is not in the model_fields
        """
        getattr(self, link_type).remove(link)


class Info(BaseModel):
    ampliconsize: PositiveInt
    schemeversion: Annotated[str, AfterValidator(validate_schemeversion)]
    schemename: Annotated[str, AfterValidator(validate_schemename)]
    primer_bed_md5: str
    reference_fasta_md5: str
    status: SchemeStatus
    citations: set[str]
    authors: Annotated[list[str], AfterValidator(not_empty)]
    algorithmversion: str
    species: Annotated[set[int | str], AfterValidator(not_empty)]
    license: str = "CC BY-SA 4.0"
    primerclass: PrimerClass = PrimerClass.PRIMERSCHEMES
    infoschema: str = INFO_SCHEMA
    articbedversion: BedfileVersion
    collections: set[Collection] = set()
    links: Links = Links()
    refselect: dict[str, dict[str, str]] | None = None

    # Add the optional fields
    description: str | None = None
    derivedfrom: str | None = None
    contactinfo: str | None = None

    def add_refselect(
        self, ref_select_file: pathlib.Path, chrom: str, info_path: pathlib.Path
    ):
        """
        Add a reference selection file to the ref select dictionary
        """
        if "-" in self.schemeversion:
            raise ValueError(
                "Cannot add reference selection to a version with a mapping suffix"
            )

        # Validate the reference selection file
        validate_ref_select_file(self, chrom, ref_select_file, info_path)

        # Read the ref selection into memory
        new_ref_select_path = info_path.parent / f"{chrom}_refselect.fasta"

        ref_select = list(SeqIO.parse(ref_select_file, "fasta"))
        with open(new_ref_select_path, "w") as f:
            SeqIO.write(ref_select, f, "fasta")

        # Add the reference selection file
        if self.refselect is None:
            self.refselect = {}
        self.refselect[chrom] = {
            "md5": hash_file(new_ref_select_path),
            "filename": new_ref_select_path.name,
        }  # type: ignore

    def collection_add(self, collection: Collection):
        """Add a collection to the collections set"""
        self.collections.add(collection)

    def collection_remove(self, collection: Collection):
        """
        Remove a collection from the collections set.
        Raises: KeyError if the collection is not in the set
        """
        self.collections.remove(collection)

    def citation_add(self, citation: str):
        """Add a citation to the citations set"""
        self.citations.add(citation)

    def citation_remove(self, citation: str):
        """
        Remove a citation from the citations set.
        Raises: KeyError if the citation is not in the set
        """
        self.citations.remove(citation)

    def author_add(self, author: str, author_index: int | None):
        """Add an author to the authors list"""
        if author not in self.authors:
            if author_index is None:
                self.authors.append(author)
            else:
                # Insert is safe, will append if index is out of range
                self.authors.insert(author_index, author)

    def author_remove(self, author: str):
        """
        Remove an author from the authors list.
        Raises: ValueError if the author is not in the list
        """
        self.authors.remove(author)

    def author_reorders(self, new_order: list[int]):
        """
        Reorder the authors list
        Raises:
            ValueError if duplicate indexes are found
            IndexError if the index is out of range
        """
        if len(new_order) != len(set(new_order)):
            raise ValueError("Duplicate indexes found")
            # Append authors in the new order

        new_authors = []
        for new_index in new_order:
            if new_index >= len(self.authors) or new_index < 0:
                raise IndexError(f"{new_index} is out of range")
            new_authors.append(self.authors[new_index])

        # Append any authors not in the new order
        for index, author in enumerate(self.authors):
            if index not in new_order:
                new_authors.append(author)

        self.authors = new_authors

    def change_primerclass(self, new_class: PrimerClass):
        """Change the primerclass"""
        self.primerclass = new_class

    def change_status(self, new_status: SchemeStatus):
        """Change the status"""
        self.status = new_status

    def change_license(self, new_license: str):
        """Change the license"""
        self.license = new_license

    def change_derivedfrom(self, new_derivedfrom: str):
        """Change the derivedfrom"""
        self.derivedfrom = new_derivedfrom

    def change_description(self, new_description: str):
        """Change the description"""
        self.description = new_description

    def change_contactinfo(self, new_contactinfo: str | None):
        """Change the contactinfo"""
        self.contactinfo = new_contactinfo

    def get_schemepath(self) -> str:
        return f"{self.schemename}/{self.ampliconsize}/{self.schemeversion}"


def validate_ref_select_file(
    info: Info, chrom: str, ref_select: pathlib.Path, infopath: pathlib.Path
):
    """
    Validate the reference selection file
    """
    # Read in the bedfile
    _header, bedlines = BedLineParser.from_file(infopath.parent / "primer.bed")
    bedline_chrom_names = {bedline.chrom for bedline in bedlines}

    # Check the chrom is in the bedfile
    if chrom not in bedline_chrom_names:
        raise ValueError(f"Chromosome ({chrom}) not in primer.bed")

    # Check the file exists
    if not ref_select.is_file():
        raise FileNotFoundError(f"{ref_select} is not a file")

    # Check the file it self
    ref_select_index = SeqIO.index(str(ref_select), "fasta")
    # Check genomes len
    ref_select_lengths = {len(seq) for seq in ref_select_index.values()}
    if len(ref_select_lengths) != 1:
        raise ValueError(
            f"Chrom ({chrom}) in {infopath}:{ref_select.name} has different lengths: {ref_select_lengths}"
        )
    # Check chrom inside fasta
    if chrom not in ref_select_index:
        raise ValueError(f"Chrom ({chrom}) not in {infopath}:{ref_select.name}")

    # Check the fasta headers
    for record in ref_select_index.values():
        if len(record.description.split()) != 2:
            raise ValueError(
                f"Chrom {chrom} in {infopath}:{ref_select.name} has fasta headers without clade tags"
            )

    return


if __name__ == "__main__":
    info = Info(
        ampliconsize=400,
        schemeversion="v0.0.0",
        schemename="test",
        primer_bed_md5="hello",
        reference_fasta_md5="world",
        status=SchemeStatus.DRAFT,
        citations=set(),
        authors=["artic"],
        algorithmversion="test",
        species=set(["sars-cov-2"]),
        articbedversion=BedfileVersion.V3,
        collections=set(),
        links=Links(),
    )

    info_json = info.model_dump_json()



import hashlib
import json
import pathlib
from typing import Annotated

import typer
from Bio import SeqIO
from primalbedtools.bedfiles import BedFileModifier, BedLineParser

from primal_page.bedfiles import BedfileVersion
from primal_page.logging import log
from primal_page.modify import generate_files, hash_file
from primal_page.schemas import INFO_SCHEMA, Info

app = typer.Typer(no_args_is_help=True)


@app.command(no_args_is_help=True)
def regenerate(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json",
            readable=True,
            exists=True,
            dir_okay=False,
            writable=True,
        ),
    ],
):
    """
    Regenerate the info.json and README.md file for a scheme
        - Rehashes info.json's primer_bed_md5 and reference_fasta_md5
        - Regenerates the README.md file
        - Recalculate the artic-primerbed version
        - Updates the infoschema version to current

    Ensures work/config.json has no absolute paths
        - Ensures hashes in config.json are removed
    """
    # Check that this is an info.json file (for safety)
    if schemeinfo.name != "info.json":
        raise typer.BadParameter(f"{schemeinfo} is not an info.json file")

    # Get the scheme path
    scheme_path = schemeinfo.parent

    # Get the info
    info_json = json.load(schemeinfo.open())

    info = Info(**info_json)
    info.infoschema = INFO_SCHEMA

    # get scheme id
    scheme_id = f"{info.schemename}/{info.ampliconsize}/{info.schemeversion}"

    # Trim whitespace from primer.bed and reference.fasta
    headers, bedlines = BedLineParser().from_file(scheme_path / "primer.bed")
    bedlines = BedFileModifier.sort_bedlines(bedlines)
    bedlines = BedFileModifier.update_primernames(bedlines)

    # If the hash is different, rewrite the file
    bedfile_str = BedLineParser().to_str(headers, bedlines)
    if (
        hash_file(scheme_path / "primer.bed")
        != hashlib.md5(bedfile_str.encode()).hexdigest()
    ):
        log.info(f"Regenerating primer.bed for {scheme_id}")
        BedLineParser().to_file(scheme_path / "primer.bed", headers, bedlines)

    # Hash the reference.fasta file
    # If the hash is different, rewrite the file
    ref_hash = hash_file(scheme_path / "reference.fasta")
    ref_str = "".join(
        x.format("fasta") for x in SeqIO.parse(scheme_path / "reference.fasta", "fasta")
    )
    if ref_hash != hashlib.md5(ref_str.encode()).hexdigest():
        log.info(f"Regenerating reference.fasta for {scheme_id}")
        with open(scheme_path / "reference.fasta", "w") as ref_file:
            ref_file.write(ref_str)

    # if articbedversion not set then set it
    articbedversion = BedfileVersion.V3
    if articbedversion == BedfileVersion.INVALID:
        raise typer.BadParameter(
            f"Could not determine artic-primerbed version for {scheme_path / 'primer.bed'}"
        )
    info.articbedversion = articbedversion

    # Regenerate the files hashes
    info.primer_bed_md5 = hash_file(scheme_path / "primer.bed")
    info.reference_fasta_md5 = hash_file(scheme_path / "reference.fasta")

    #####################################
    # Final validation and create files #
    #####################################

    pngs = list(schemeinfo.parent.rglob("*.png"))
    generate_files(info, schemeinfo, pngs)


@app.command(no_args_is_help=True)
def regenerate_all(
    primerschemes: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The parent directory",
            readable=True,
            exists=True,
            writable=True,
            file_okay=False,
        ),
    ],
):
    """
    THIS MODIFIES THE SCHEMES IN PLACE. USE WITH CAUTION
        Regenerates all schemes in the primerschemes directory.
        Mainly used for migrating to the new info.json schema.
    """
    # Get all the schemes
    info_jsons = list(primerschemes.rglob("info.json"))

    for info_json in info_jsons:
        regenerate(info_json)



from enum import Enum

# Bedfile versions
## This doesn't parse the contents just the structure
BEDFILE_LINE = r"^\S+\t\d+\t\d+\t\S+\t\d+\t(\+|\-)\t[a-zA-Z]+$"


# bedfile versions
class BedfileVersion(Enum):
    """
    V1 bedfiles use a 6 col system
    V2 bedfiles use a 7 col system and V1 primernames
    V3 bedfiles use a 7 col system and V2 primernames
    """

    V1 = "v1.0"
    V2 = "v2.0"
    V3 = "v3.0"
    INVALID = "invalid"  # Not applicable



from importlib.metadata import version

from typer.testing import CliRunner

from primal_page.main import app

runner = CliRunner()

# This uses pytest rather than unittest
# Run using: poetry run pytest


# Test the app can run with the version flag
def test_app_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert version("primal-page") in result.stdout



import unittest

from primal_page.errors import (
    InvalidSchemeID,
    InvalidSchemeName,
    InvalidSchemeVersion,
)
from primal_page.schemas import (
    not_empty,
    validate_scheme_id,
    validate_schemename,
    validate_schemeversion,
)


class TestValidateSchemeVersion(unittest.TestCase):
    def test_validate_schemeversion_valid(self):
        """
        Tests main/VERSION_PATTERN for valid scheme names
        """
        valid_versions = [
            "v1.0.0",
            "v2.3.4",
            "v10.20.30",
            "v1.0.0-beta",
            "v1.0.0-alpha",
            "v1.0.0-alpha1",
        ]

        for version in valid_versions:
            self.assertEqual(validate_schemeversion(version), version)

    def test_validate_schemeversion_invalid(self):
        """
        Tests main/VERSION_PATTERN for invalid scheme names
        """
        invalid_versions = [
            "v1",
            "v2.3",
            "v10.20",
            "v1.0.0.0",
            "V1",
            "artic-v1.0.0",
            "v1.0.0-",
            "v1.0.0-!",
            "v1.0.0-alpha!",
            "v1.0.0-A",
            "v1.0.0_alpha",
            "v1.0.0.alpha",
        ]

        for version in invalid_versions:
            with self.assertRaises(InvalidSchemeVersion):
                validate_schemeversion(version)


class TestRegex(unittest.TestCase):
    def test_SchemeNamePattern_valid(self):
        """
        Tests main/SCHEMENAME_PATTERN for valid scheme names
        """
        valid_names = {
            "artic-covid-400": "artic-covid-400",
            "mpx-2022": "mpx-2022",
            "midnight-bccdc-2021": "midnight-bccdc-2021",
            "primerscheme": "primerscheme",
        }

        for name, result in valid_names.items():
            self.assertEqual(validate_schemename(name), result)

    def test_SchemeNamePattern_invalid(self):
        """
        Tests main/SCHEMENAME_PATTERN for invalid scheme names
        """

        invalid_names = [
            "artic-covid-400-",
            "artic/covid-400-1",
            "artic-covid-400-1.0!",
            "*artic-covid-400-1.0",
        ]

        for name in invalid_names:
            with self.assertRaises(InvalidSchemeName):
                validate_schemename(name)


class TestNotEmpty(unittest.TestCase):
    def test_not_empty_full(self):
        test_cases = [[1], {1}, [1, 2], {10, 11}]
        for test_case in test_cases:
            self.assertEqual(not_empty(test_case), test_case)

    def test_not_empty_empty(self):
        test_cases = [[], {}, set(), ""]
        for test_case in test_cases:
            with self.assertRaises(ValueError):
                not_empty(test_case)


class TestValidateSchemeID(unittest.TestCase):
    def test_valid_scheme_id(self):
        valid_scheme_id = ["test123/400/v1.0.0", "test123/1000/v1.0.0"]

        # Valid scheme ids
        for scheme_id in valid_scheme_id:
            self.assertEqual(tuple(scheme_id.split("/")), validate_scheme_id(scheme_id))

    def test_invalid_scheme_id_schemename(self):
        # Scheme ids with invalid scheme names
        invalid_scheme_id = ["test123-/400/v1.0.0", "test1A23/1000/v1.0.0"]

        # Invalid scheme ids
        for scheme_id in invalid_scheme_id:
            with self.assertRaises(InvalidSchemeID):
                validate_scheme_id(scheme_id)

    def test_invalid_scheme_id_ampliconsize(self):
        # Scheme ids with invalid ampliconsize
        invalid_scheme_id = ["test123/40a0/v1.0.0", "test123/10.00/v1.0.0"]

        # Invalid scheme ids
        for scheme_id in invalid_scheme_id:
            with self.assertRaises(InvalidSchemeID):
                validate_scheme_id(scheme_id)

    def test_invalid_scheme_id_schemeversion(self):
        # Scheme ids with invalid ampliconsize
        invalid_scheme_id = ["test123/400/v1", "test123/1000/1.0"]

        # Invalid scheme ids
        for scheme_id in invalid_scheme_id:
            with self.assertRaises(InvalidSchemeID):
                validate_scheme_id(scheme_id)

    def test_invalid_scheme_id_structure(self):
        # Scheme ids with invalid structure
        invalid_scheme_id = ["test123/400", "test123/1000/1.0/test"]

        # Invalid scheme ids
        for scheme_id in invalid_scheme_id:
            with self.assertRaises(InvalidSchemeID):
                validate_scheme_id(scheme_id)


if __name__ == "__main__":
    unittest.main()
